import json
import os
import platform
import re
import shutil
import sys
import tempfile
import threading
from subprocess import PIPE, Popen
from functools import reduce

from demogpt.chains.task_chains import TaskChains
from demogpt.controllers import checkPromptTemplates, refineKeyTypeCompatiblity

AI_VARIETY_TEMPERATURE = 0.5

def init(title=""):
    initial_code = IMPORTS_CODE_SNIPPET 
    if title:
        initial_code += f"\nst.title('{title}')\n"
    return initial_code

def filterTasks(tasks):
    filtered_tasks = []
    for task in tasks:
        if len(task["input_key"]) == 0 and len(task["output_key"]) == 0:
            ...
        else:
            filtered_tasks.append(task)
    return filtered_tasks

def reorderTasksForChatApp(tasks):
    print("before reordering:")
    print(tasks)
    step2task={task["step"]:task for task in tasks}
    input_chat_step = -1
    for task in tasks:
        if task["task_type"] == "ui_input_chat":
            input_chat_step = task["step"]
            break
    
    if input_chat_step == -1:
        return tasks 

    input2tasks={}
    for task in tasks:
        for input_key in task["input_key"]:
            input2tasks[input_key] = input2tasks.get(input_key,[]) + [task]
    
    non_pre_task_steps = set() # these are the ones that cannot be done before chat input
    steps = {i for i in range(1,len(tasks)+1)}
    all_steps = {i for i in range(1,len(tasks)+1)}
    root = step2task[input_chat_step]
    non_pre_task_steps.add(root["step"])
    change = True
    while change:
        change = False
        for step in non_pre_task_steps:
            task1 = step2task[step]
            for input_key in task1["input_key"]:
                tasks1 = input2tasks[input_key]
                for task2 in tasks1:
                    if task2["step"] not in non_pre_task_steps:
                        change = True
                        non_pre_task_steps.add(task2["step"])
                        steps.remove(task["step"])
    post_steps = sorted(list(all_steps-steps))
    steps = sorted(list(steps))     
    pre_tasks = [step2task[step] for step in steps]
    post_tasks = [step2task[step] for step in post_steps]
    new_tasks = pre_tasks + post_tasks
    print("after reordering:")
    print(new_tasks)      
    return new_tasks

def getFunctionNames(code):
    pattern = r"def (\w+)\(.*\):"
    return re.findall(pattern, code)


def getGenericPromptTemplateCode(task, iters):
    res = ""
    is_valid = False
    task_type = task["task_type"]
    inputs = task["input_key"]
    prompt_func = (
        TaskChains.promptTemplate if task_type == "prompt_template" else TaskChains.chat
    )
    finalizer_func = (
        getPromptChatTemplateCode if task_type == "prompt_template" else getChatCode
    )
    additional_inputs = []
    if task_type == "chat":
        additional_inputs.append("chat_history")
    res = prompt_func(task=task)
    function_name = res.get("function_name")
    variety = res.get("variety")
    index = 0
    while not is_valid:
        templates = {key: res.get(key) for key in res if "template" in key}
        check = checkPromptTemplates(templates, task, additional_inputs)
        is_valid = check["valid"]
        feedback = check["feedback"]
        if not is_valid:
            res = TaskChains.promptTemplateRefiner(res, inputs, feedback)
        else:
            break
        index += 1
        if index == iters:
            break
    res["function_name"] = function_name
    res["variety"] = variety
    return finalizer_func(res, task)


def getCodeSnippet(task, code_snippets, iters=10):
    #task = refineKeyTypeCompatiblity(task)
    task_type = task["task_type"]
    code = ""
    if task_type == "ui_input_text":
        code = TaskChains.uiInputText(task=task)
    elif task_type == "ui_output_text":
        code = TaskChains.uiOutputText(task=task)
    elif task_type in ["prompt_template", "chat"]:
        code = getGenericPromptTemplateCode(task, iters=iters)
    elif task_type == "path_to_content":
        code = TaskChains.pathToContent(task=task, code_snippets=code_snippets)
    elif task_type == "doc_to_string":
        code = TaskChains.docToString(task=task)
    elif task_type == "string_to_doc":
        code = TaskChains.stringToDoc(task=task)
    elif task_type == "ui_input_file":
        code = TaskChains.uiInputFile(task=task)
    elif task_type == "doc_loader":
        code = TaskChains.docLoad(task=task, code_snippets=code_snippets)
    elif task_type == "doc_summarizer":
        code = TaskChains.summarize(task=task)
    elif task_type == "ui_input_chat":
        code = getChatInputCode(TaskChains.uiInputChat(task=task))
    elif task_type == "ui_output_chat":
        code = TaskChains.uiOutputChat(task=task)
    elif task_type == "python":
        code = TaskChains.pythonCoder(task=task, code_snippets=code_snippets)
    elif task_type == "plan_and_execute":
        code = TaskChains.search(task=task)
    elif task_type == "search_chat":
        code = TaskChains.search_chat(task=task)
    return code.strip() + "\n"


def getChatInputCode(code):
    prefix = """
for message in st.session_state.messages:
    with st.chat_message(message["role"]):  
        st.markdown(message["content"])\n
"""
    return prefix + code


def refine(code):
    if "```" in code:
        code = code.split("```")[1]
        if code.startswith("python"):
            code = code[len("python") :].strip()
    return code


def reformatTasks(tasks):
    def proprocess(io):
        if io == "none":
            io = []
        elif isinstance(io, str):
            if io.startswith("["):
                io = io[1:-1]
            io = [var.strip() for var in io.split(",")]
        return io
    
    tasks = filterTasks(tasks)

    processed_tasks = []
    for task in tasks:
        task_input = proprocess(task["input_key"])
        task_output = proprocess(task["output_key"])
        task["input_key"] = task_input
        task["output_key"] = task_output
        processed_tasks.append(task)

    return processed_tasks


def getChatCode(template, task):
    def getHumanInput(template, inputs):
        rightmost = -1
        human_input = ""
        for input in inputs:
            pattern = f"{{{input}}}"
            index = template.rfind(pattern)
            if index > rightmost:
                human_input = input
                rightmost = index
        return human_input

    inputs = task["input_key"]
    variable = ", ".join(task["output_key"])
    temperature = 0 if template.get("variety", "False") == "False" else AI_VARIETY_TEMPERATURE
    system_template = template["system_template"]
    run_call = "{}"

    if inputs == "none":
        signature = f"{template['function_name']}()"
        function_call = f"{variable} = {signature}"
        inputs = []
    else:
        if isinstance(inputs, str):
            if inputs.startswith("["):
                inputs = inputs[1:-1]
            inputs = [var.strip() for var in inputs.split(",")]
        if len(inputs) > 0:
            run_call = ", ".join([f"{var}={var}" for var in inputs])
        signature = f"{template['function_name']}({','.join(inputs)})"
        function_call = f"""
if not openai_api_key.startswith('sk-'):
    st.warning('Please enter your OpenAI API key!', icon='⚠')
    {variable} = ""
elif {' and '.join(inputs)}:
    with st.spinner('DemoGPT is working on it. It takes less than 10 seconds...'):
        {variable} = chat_llm_chain.run({run_call})
else:
    {variable} = ""
"""
    input_variables = ["chat_history"] + inputs
    human_input = template["human_input"]
    if human_input not in inputs:
        human_input = getHumanInput(system_template, inputs)
    code = f"""
from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate
from langchain.memory.chat_message_histories import StreamlitChatMessageHistory
from langchain.memory import ConversationBufferMemory
from langchain.chat_models import ChatOpenAI

msgs = StreamlitChatMessageHistory()

prompt = PromptTemplate(
    input_variables={input_variables}, template='''{system_template}'''
)
memory = ConversationBufferMemory(memory_key="chat_history", input_key="{human_input}", chat_memory=msgs, return_messages=True)
llm = ChatOpenAI(model_name="gpt-3.5-turbo-16k", openai_api_key=openai_api_key, temperature={temperature})
chat_llm_chain = LLMChain(
    llm=llm,
    prompt=prompt,
    verbose=False,
    memory=memory,
)
    
{function_call} 

    """

    return code


def getPromptChatTemplateCode(templates, task):
    inputs = task["input_key"]
    variable = ", ".join(task["output_key"])
    run_call = "{}"

    if inputs == "none":
        signature = f"{templates['function_name']}()"
        function_call = f"{variable} = {signature}"
    else:
        if isinstance(inputs, str):
            if inputs.startswith("["):
                inputs = inputs[1:-1]
            inputs = [var.strip() for var in inputs.split(",")]
        if len(inputs) > 0:
            run_call = ", ".join([f"{var}={var}" for var in inputs])
        signature = f"{templates['function_name']}({','.join(inputs)})"
        function_call = f"""
if not openai_api_key.startswith('sk-'):
    st.warning('Please enter your OpenAI API key!', icon='⚠')
    {variable} = ""
elif {' and '.join(inputs)}:
    with st.spinner('DemoGPT is working on it. It takes less than 10 seconds...'):
        {variable} = {signature}
else:
    {variable} = ""
"""

    temperature = 0 if templates.get("variety", "False") == "False" else AI_VARIETY_TEMPERATURE

    code = f"""\n
from langchain.chains import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (ChatPromptTemplate, HumanMessagePromptTemplate, SystemMessagePromptTemplate)

def {signature}:
    chat = ChatOpenAI(
        model="gpt-3.5-turbo-16k",
        openai_api_key=openai_api_key,
        temperature={temperature}
    )
    system_template = \"\"\"{templates['system_template']}\"\"\"
    system_message_prompt = SystemMessagePromptTemplate.from_template(system_template)
    human_template = \"\"\"{templates['template']}\"\"\"
    human_message_prompt = HumanMessagePromptTemplate.from_template(human_template)
    chat_prompt = ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )

    chain = LLMChain(llm=chat, prompt=chat_prompt)
    result = chain.run({run_call})
    return result # returns string   

{function_call}               

"""
    return code


def runThread(proc):
    proc.communicate()


def runStreamlit(code, openai_api_key, openai_api_base=None):
    """
    Runs the provided code as a Streamlit application and returns the process ID.

    Args:
        code (str): The code of the Streamlit application.

    Returns:
        int: The process ID of the Streamlit application.
    """
    tmp = tempfile.NamedTemporaryFile("w", suffix=".py", delete=False, encoding="utf-8")
    tmp.write(code)
    tmp.flush()
    environmental_variables = {
        "OPENAI_API_KEY": openai_api_key,
        "STREAMLIT_SERVER_PORT": "8502",
        "OPENAI_API_BASE": openai_api_base,
    }
    streamlit_path = shutil.which("streamlit")
    if True or platform.system() == "Windows":
        env = os.environ.copy()
        env["PYTHONPATH"] = ""
        env["OPENAI_API_KEY"] = openai_api_key
        env["STREAMLIT_SERVER_PORT"] = "8502"
        if openai_api_base:
            env["OPENAI_API_BASE"] = openai_api_base
        python_path = sys.executable
        process = Popen(
            [python_path, "-m", "streamlit", "run", tmp.name],
            env=env,
            stdout=PIPE,
            stderr=PIPE,
        )
        threading.Thread(target=runThread, args=(process,)).start()
    try:
        tmp.close()
    except PermissionError:
        pass

    return process.pid


IMPORTS_CODE_SNIPPET = """
import os
import streamlit as st
import tempfile

# Initialize chat history
if "messages" not in st.session_state:
    st.session_state.messages = []

openai_api_key = st.sidebar.text_input(
    "OpenAI API Key",
    placeholder="sk-...",
    value=os.getenv("OPENAI_API_KEY", ""),
    type="password",
)
"""



# human_template = """
# Write a streamlit textfield or textarea code depending on the instruction below:
# Suppose that, streamlit has been imported by "import streamlit as st" so you don't need to import it.
# Here is the part of the code that you are supposed to continue:
# {code_snippets}
#
# Assign the taken input to the variable called "{variable}"
#
# Instruction:{instruction}
# Streamlit Code:
# """

human_template = """
variable: url
instruction: Get website URL from the user
code:
url = st.text_input("Enter website URL")

variable: source_text
instruction: Get source text from the user
code:
source_text = st.text_area("Enter source text")

variable: input_language
instruction: Get the input language from the user
code:
input_language = st.text_input("Enter the input language")

variable: color
instruction: Select the color
code:
color = st.selectbox("Select the color", ["Red", "Blue", "Green", "Yellow", "purple"])

variable: {variable}
instruction: {instruction}
code:

"""



system_template = """
Create a new plan which does the given instruction that does not have the problems of the old plan

Each step can only use one of the functions below:

{TASK_NAMES}

These are the explanations of those functions:

{TASK_PURPOSES}

The plan should be in the same format as the Problematic Plan but it cannot include the problem in the "Problems" section.

It is extremely important that all reasoning steps should be in the following format:
$step_num. $description [$task_type($arguments) ---> $result]

$step_num: is positive integer showing the order of the reasoning step.
$description : says the responsibility of the task.
$task_type : one of the available functions in {TASK_NAMES}
$arguments: input variable(s) for the function (it should be output of the one of the previous functions)
$result : output variable for the function
"""

human_template = """
Instruction:{instruction}

Problematic Plan:
{plan}

Problems: {feedback}

New Refined Plan:
"""



import re
from functools import reduce
from demogpt.chains.task_definitions import getTasks

def planToTaskFormat(plan):
    pattern = re.compile(r"\b(\w+)\(([^)]*)\)( ---> ([\w_,\s]+))?")
    matches = pattern.findall(plan)
    functions = [
        {
            "name": match[0],
            "arguments": [
                arg.strip()
                for arg in match[1].split(",")
                if arg.strip() and '"' not in arg.strip()
            ],
            "output": [
                arg.strip()
                for arg in match[3].split(",")
                if arg.strip() and '"' not in arg.strip()
            ],
        }
        for match in matches
    ]
    plan_steps = []
    for i, function in enumerate(functions):
        step = {
            "step": i + 1,
            "task_type": function["name"],
            "task_name": function["name"],
            "input_key": function["arguments"],
            "output_key": function["output"],
        }
        plan_steps.append(step)
    return plan_steps

def checkTaskNames(tasks, app_type):
    TASK_TYPE2_TASK = getTasks(app_type)[-1]
    feedback = ""
    for task in tasks:
        name = task["task_type"]
        if name not in TASK_TYPE2_TASK:
            feedback += f"Task {name} is invalid, you cannot use undefined task.\n"
            
    if feedback:
        feedback += "\nOnly those tasks are available:" + " ".join(
                TASK_TYPE2_TASK.keys()
            )
    
    valid = len(feedback) == 0
    
    return {
        "feedback": feedback,
        "valid": valid
    }

def validate(input, app_type):
    if isinstance(input,str):#means plan
        input = planToTaskFormat(input) 
    
    res1 = checkTaskNames(input, app_type)
    res2 = checkAppTypeCompatiblity(input, app_type)
    res3 = checkRedundantTasks(input)
    res4 = checkInputOutputLengthCompatiblity(input, app_type)
    res5 = checkInputOuputCompatibility(input)

    feedback = ""
    
    if not res1["valid"]:
        feedback += "\n" + res1["feedback"]
    if not res2["valid"]:
        feedback += "\n" + res2["feedback"]
    if not res3["valid"]:
        feedback += "\n" + res3["feedback"]
    if not res4["valid"]:
        feedback += "\n" + res4["feedback"]
    if not res5["valid"]:
        feedback += "\n" + res5["feedback"]

    valid = len(feedback) == 0
    
    print("feedback:")
    print(feedback)

    return {"feedback": feedback, "valid": valid}

def refineKeyTypeCompatiblity(task):
    if task["input_data_type"] == "none":
        if task["input_key"] != "none":
            task["input_key"] = "none"
    if task["output_data_type"] == "none":
        if task["output_key"] != "none":
            task["output_key"] = "none"
    return task


def checkInputOuputCompatibility(tasks):
    feedback = ""
    inputs = set((reduce((lambda x, y: x+y),[task["input_key"] for task in tasks] + [["none"]])))
    outputs = set((reduce((lambda x, y: x+y),[task["output_key"] for task in tasks] + [["none"]])))
        
    input2tasks = {}
    output2tasks = {}
    for task in tasks:
        for input_key in task["input_key"]:
            input_task_names = input2tasks.get(input_key,[])
            input_task_names.append(task["task_name"])
            input2tasks[input_key] = input_task_names
        for output_key in task["output_key"]:
            output_task_names = output2tasks.get(output_key,[])
            output_task_names.append(task["task_name"])
            output2tasks[output_key] = output_task_names
    
    ghost_inputs = inputs - outputs # inputs which are not coming from output of another task
    redundant_outputs = outputs - inputs # outputs which are an input of another task
    
    for input_key in ghost_inputs:
        for task_name in input2tasks[input_key]:
            feedback += f"Remove task {task_name} because its input {input_key} is not coming from another task.\n" 
    for output_key in redundant_outputs:
        for task_name in output2tasks[output_key]:
            feedback += f"Remove task {task_name} because its output {output_key} is not used. It is redundant.\n" 
    
    valid = len(feedback) == 0
    
    return {
        "valid": valid,
        "feedback": feedback
    }
        

def checkRedundantTasks(tasks):
    all_input_keys = []
    for task in tasks:
        inputs = task["input_key"]
        all_input_keys += inputs

    all_input_keys = set(all_input_keys)
    all_input_keys.add("none")

    feedback = ""

    for task in tasks:
        for output_key in task["output_key"]:
            if output_key not in all_input_keys:
                feedback += f"Task {task['task_name']} has problem because its output {output_key} is never used. Try another approach.\n"

    valid = len(feedback) == 0

    return {"feedback": feedback, "valid": valid}

def checkInputOutputLengthCompatiblity(tasks, app_type):
    TASK_TYPE2_TASK = getTasks(app_type)[-1]
    feedback = ""
    for task in tasks:
        task_type = task["task_type"]
        if task_type not in TASK_TYPE2_TASK:
            continue
        
        original_task = TASK_TYPE2_TASK[task_type]
        # input check
        if original_task["input_data_type"] == "none":
            if len(task["input_key"]) > 0:
                feedback += f"Task {task_type} cannot have input\n"
        
        elif not original_task["input_data_type"].startswith("*"):
            if len(task["input_key"]) > 1:
                feedback += f"Task {task_type} can only have single input but you gave multiple"
                
        # output check
        if original_task["output_data_type"] == "none":
            if len(task["output_key"]) > 0:
                feedback += f"Task {task_type} cannot have output\n"
                
        elif not original_task["output_data_type"].startswith("*"):
            if len(task["output_key"]) > 1:
                feedback += f"Task {task_type} can only have single output but you gave multiple"
                
    valid = len(feedback) == 0
    
    return {
        "feedback": feedback,
        "valid": valid
    }
    

def checkAppTypeCompatiblity_old(tasks, app_type):
    must_chat_tasks = {"ui_input_chat", "chat", "ui_output_chat"}
    must_prompt_template_tasks = {"prompt_template"}
    must_search_tasks = {"plan_and_execute"}

    task_types_list= [task["task_type"] for task in tasks]
    task_types = set(task_types_list)

    task_prompt_template = must_prompt_template_tasks & task_types
    task_search = must_search_tasks & task_types
    task_chat = must_chat_tasks & task_types

    app_prompt_template = 0  # neutral

    app_chat = app_type["is_chat"] == "true"
    app_search = app_type["is_search"] == "true"
    if not app_chat:
        if app_type["is_ai"] == "true":
            app_prompt_template = 1
        else:
            app_prompt_template = -1

    feedback = ""
    ################################################################################################################################################
    # chat app check
    if app_chat:
        for task_type in must_chat_tasks - task_chat:
            feedback += f"The app is chat-based but you didn't use {task_type} in your tasks. Please add it and try again\n"
            
        for task_type in must_chat_tasks:
            if task_types_list.count(task_type) > 1:
                feedback += f"You can use {task_type} in your tasks only once. Please remove the redundant ones and combine in a single task\n"
    else:
        for task_type in task_chat:
            feedback += f"The app is not chat-based but you used {task_type} in your task list. Please remove it and try again\n"
    ################################################################################################################################################
    # search app check
    if app_search:
        for task_type in must_search_tasks - task_search:
            feedback += f"The app requires {task_type} task but you didn't use {task_type} in your tasks. Please add it and try again\n"
    else:
        for task_type in task_search:
            feedback += f"The app does not need {task_type} task but you used {task_type} task in your task list. Please remove them and try again\n"
    ################################################################################################################################################
    # prompt_template app check
    if app_prompt_template == 1:
        for task_type in must_prompt_template_tasks - task_prompt_template:
            feedback += f"The app is ai-based but you didn't use {task_type} in your tasks. Please add it and try again\n"
    elif app_prompt_template == -1:
        for task_type in task_prompt_template:
            feedback += f"The app is not ai-based but you used {task_type} in your tasks which is redundant. Please remove it and try again\n"
    ################################################################################################################################################
    # search-python compatiblity
    python_tasks = [task for task in tasks if task["task_type"] == "python"]
    search_tasks = [task for task in tasks if task["task_type"] == "plan_and_execute"]
    found = False
    for python_task in python_tasks:
        python_inputs = set(python_task["input_key"])
        for search_task in search_tasks:
            search_outputs = set(search_task["output_key"])
            if python_inputs & search_outputs:
                feedback += f""" python task '{python_task['task_name']}' uses {(python_inputs & search_outputs).pop()} as an input but it comes from plan_and_execute task '{search_task['task_name']}'. python task cannot use plan_and_execute task's output as an input. Please redesign the tasks so that no python task uses plan_and_execute task's output as an input!"""
                found = True
                break
        if found:
            break

    ################################################################################################################################################
    # plan_and_execute compatibility
    search_indices = sorted([task["step"] for task in search_tasks])
    if len(search_tasks) > 1:
        for index in search_indices:
            if index + 1 in search_indices:
                feedback += """It is not recommended to use back to back "plan_and_execute" tasks because one plan_and_execute can handle generic question by itself. 
                You should combine plan_and_execute tasks in a single task."""
                break

    valid = len(feedback) == 0

    return {"feedback": feedback, "valid": valid}


def checkAppTypeCompatiblity(tasks, app_type):
    must_chat_tasks = {"ui_input_chat", "chat", "ui_output_chat", "search_chat"}
    must_prompt_template_tasks = {"prompt_template"}
    must_search_tasks = {"plan_and_execute", "search_chat"}

    task_types_list= [task["task_type"] for task in tasks]
    task_types = set(task_types_list)

    app_prompt_template = 0  # neutral

    app_chat = app_type["is_chat"] == "true"
    app_search = app_type["is_search"] == "true"
    if not app_chat:
        if app_type["is_ai"] == "true":
            app_prompt_template = 1
        else:
            app_prompt_template = -1
            
    if app_chat:
        if app_search:
            must_search_tasks.remove("plan_and_execute")
            must_chat_tasks.remove("chat")
    else:
        must_search_tasks.remove("search_chat")
        must_chat_tasks.remove("search_chat")
        
    task_prompt_template = must_prompt_template_tasks & task_types
    task_search = must_search_tasks & task_types
    task_chat = must_chat_tasks & task_types        

    feedback = ""
    ################################################################################################################################################
    # chat app check
    if app_chat:
        for task_type in must_chat_tasks - task_chat:
            feedback += f"The app is chat-based but you didn't use {task_type} in your tasks. Please add it and try again\n"
            
        for task_type in must_chat_tasks:
            if task_types_list.count(task_type) > 1:
                feedback += f"You can use {task_type} in your tasks only once. Please remove the redundant ones and combine in a single task\n"
    else:
        for task_type in task_chat:
            feedback += f"The app is not chat-based but you used {task_type} in your task list. Please remove it and try again\n"
    ################################################################################################################################################
    # search app check
    if app_search:
        for task_type in must_search_tasks - task_search:
            feedback += f"The app requires {task_type} task but you didn't use {task_type} in your tasks. Please add it and try again\n"
    else:
        for task_type in task_search:
            feedback += f"The app does not need {task_type} task but you used {task_type} task in your task list. Please remove them and try again\n"
    ################################################################################################################################################
    # prompt_template app check
    if app_prompt_template == 1:
        for task_type in must_prompt_template_tasks - task_prompt_template:
            feedback += f"The app is ai-based but you didn't use {task_type} in your tasks. Please add it and try again\n"
    elif app_prompt_template == -1:
        for task_type in task_prompt_template:
            feedback += f"The app is not ai-based but you used {task_type} in your tasks which is redundant. Please remove it and try again\n"
    ################################################################################################################################################
    # search-python compatiblity
    python_tasks = [task for task in tasks if task["task_type"] == "python"]
    search_tasks = [task for task in tasks if task["task_type"] == "plan_and_execute"]
    found = False
    for python_task in python_tasks:
        python_inputs = set(python_task["input_key"])
        for search_task in search_tasks:
            search_outputs = set(search_task["output_key"])
            if python_inputs & search_outputs:
                feedback += f""" python task '{python_task['task_name']}' uses {(python_inputs & search_outputs).pop()} as an input but it comes from plan_and_execute task '{search_task['task_name']}'. python task cannot use plan_and_execute task's output as an input. Please redesign the tasks so that no python task uses plan_and_execute task's output as an input!"""
                found = True
                break
        if found:
            break

    ################################################################################################################################################
    # plan_and_execute compatibility
    search_indices = sorted([task["step"] for task in search_tasks])
    if len(search_tasks) > 1:
        for index in search_indices:
            if index + 1 in search_indices:
                feedback += """It is not recommended to use back to back "plan_and_execute" tasks because one plan_and_execute can handle generic question by itself. 
                You should combine plan_and_execute tasks in a single task."""
                break

    valid = len(feedback) == 0

    return {"feedback": feedback, "valid": valid}


def checkDTypes(tasks, app_type):
    TASK_TYPE2_TASK = getTasks(app_type)[-1]
    
    feedback = ""
    for task in tasks:
        name = task["task_type"]

        input_data_type = task["input_data_type"]
        output_data_type = task["output_data_type"]
        input_key = task["input_key"]

        if name not in TASK_TYPE2_TASK:
            feedback += (
                f"There is no task with a name {name}.Please find another way.\n"
            )
            continue

        reference = TASK_TYPE2_TASK[name]
        reference_input = reference["input_data_type"]
        reference_output = reference["output_data_type"]

        if task["step"] == 1:
            if input_key != "none":
                feedback += f"Since {name} is the first task, its input data type is supposed to be none but it is {input_key}.Please find another way.\n"

        elif reference_input == "*":
            continue

        elif reference_input.startswith("*") and input_data_type == "list":
            continue
        # Check input data types
        elif reference_input.startswith("*"):
            reference_input = reference_input.replace("*", "")
            for res, data_type in zip(input_key, input_data_type):
                if data_type != reference_input:
                    feedback += f"""
                    {name} expects all inputs as {reference_input} but data type of {res} is {data_type} not {reference_input}. Please find another way.\n
                    """
        elif input_data_type != reference_input:
            feedback += f"""
            {name} expects all inputs as {reference_input} but the data type of {input_key} is {input_data_type} not {reference_input}. Please find another way.\n
            """

        if reference_output == "*":
            continue
        # Check output data types
        elif output_data_type != reference_output:
            if not (reference_output.startswith("*") and output_data_type != "list"):
                feedback += f"""
                {name} should output in {reference_output} data type but it is {output_data_type} not {reference_output}. Please find another way.\n
                """

    valid = len(feedback) == 0

    return {"feedback": feedback, "valid": valid}


def checkPromptTemplates(templates, task, additional_inputs=[]):
    templates = " ".join(list(templates.values()))
    inputs = task["input_key"]
    if inputs == "none":
        inputs = []
    else:
        if isinstance(inputs, str):
            if inputs.startswith("["):
                inputs = inputs[1:-1]
            inputs = [var.strip() for var in inputs.split(",")]
    template_inputs = inputs + additional_inputs
    feedback = ""
    for input_key in template_inputs:
        if f"{{{input_key}}}" not in templates:
            feedback += f"'{{{input_key}}}' is not included in any of the templates. You must add '{{{input_key}}}' inside of at least one of the templates.\n"

    # now detect extras

    matches = set(re.findall(r"\{([^}]+)\}", templates))

    for match in matches:
        if match not in template_inputs:
            feedback += f"'{{{match}}}' cannot be included nowhere in the templates. You must remove '{{{match}}}'.\n"

    valid = len(feedback) == 0

    return {"feedback": feedback, "valid": valid}



system_template = """
You are a helpful assistant that can generate a concise "about" markdown for an application.
The application is generated based on the given instruction. You will also see the title of the application.
The about markdown should be short and explanatory.
"""

human_template = """
Instruction:Create an app that can answer question related to the uploaded documents
Title:📖KnowledgeGPT
About:📖KnowledgeGPT allows you to ask questions about your documents and get accurate answers with instant citations.


Instruction:{instruction}
Title:{title}
About:
"""


system_template = """
Please classify your app idea based on the following criteria and generate the appropriate JSON:

1. explanation:
    Gives detailed explanation about the decision for each one is_ai, is_chat, is_search respectively.

2. is_ai:
    - True if:
        a. The app requires natural language understanding or generation.
        b. It requires complex computations or algorithms beyond standard Python libraries like 'numpy', 'pandas', or 'requests'.
    - False otherwise.

3. is_chat:
    - True if the primary function of the app is to engage in conversation.
    - False otherwise.

4. is_search:
    - True if:
        a. The app idea strictly requires information after cut off date and this information should be fetched in the whole web instead of specific websites.
    - False otherwise.
    
Example JSON Format:

{{
    "explanation":"string",
    "is_ai":"true/false",
    "is_chat":"true/false",
    "is_search":"true/false"
}}
"""

human_template = """
App Idea: an agent that can get analysis of CSV file then summarize it.
JSON:{{
    "explanation":"Making analysis requires ai. It does not include conversation. No up to date information is needed",
    "is_ai":"true",
    "is_chat":"false",
    "is_search":"false"
}}

App Idea: an application that can get the word count of txt file.
JSON:{{
    "explanation":"Word count is a simple python task so no ai is required. No conversation is expected. No up to date information is needed",
    "is_ai":"false",
    "is_chat":"false",
    "is_search":"false"
}}

App Idea: create an application that can talk like Jeff Bezos
JSON:{{
    "explanation":"To talk like Jeff Bezos, need generative text model so I need ai. Talking app includes conversation. No up to date information is needed",
    "is_ai":"true",
    "is_chat":"true",
    "is_search":"false"
}}

App Idea: create an application that can find and list all the male names
JSON:{{
    "explanation":"I need ai to filter the male names because it is not a simple Python task. It does not include conversation. No up to date information is needed",
    "is_ai":"true",
    "is_chat":"false",
    "is_search":"false"
}}

App Idea: generate an agent that can give suggestions to the uploaded CV
JSON:{{
    "explanation":"Giving suggestions is not a simple Python task so i need ai. Giving suggestions/advice does not require a conversation. No up to date information is needed",
    "is_ai":"true",
    "is_chat":"false",
    "is_search":"false"
}}

App Idea: a system that can transform given one currency to another
JSON:{{
    "explanation": "To transform from one currency to another, i need flexible analysis so i need ai. Currency transformation does not require conversation. I need to search the up to date currency information from web",
    "is_ai":"true",
    "is_chat":"false",
    "is_search":"true"
}}

App Idea:{instruction}
JSON:
"""



import os

from langchain.chains import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (AIMessagePromptTemplate,
                                    ChatPromptTemplate,
                                    HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)
from termcolor import colored
from tqdm import trange

from . import prompts

PROMPTS = {"final": prompts.final_refiner}


class SelfRefiner:
    def __init__(
        self,
        key="final",
        max_iter=4,
        stop_kw="<SUCCESS>",
        log_intermediate_steps=True,
        model="gpt-3.5-turbo",
        openai_api_key=os.getenv("OPENAI_API_KEY", ""),
        temperature=0.0,
        openai_api_base=None,
    ):
        assert key in PROMPTS

        self.stop_kw = stop_kw
        self.max_iter = max_iter
        self.log_intermediate_steps = log_intermediate_steps

        self.prompts = {
            "feedback": PROMPTS[key].FEEDBACK_PROMPT,
            "refine": PROMPTS[key].REFINEMENT_PROMPT,
        }
        self.llm = ChatOpenAI(
            model=model,
            openai_api_key=openai_api_key,
            temperature=temperature,
            openai_api_base=openai_api_base,
        )

        self.conversation_history = self.getPromptTemplate("refine")

    def getPromptTemplate(self, key):
        assert key in self.prompts
        prompts = []
        prompts.append(SystemMessagePromptTemplate.from_template(self.prompts[key]))
        return prompts

    def addToHistory(self, prompt):
        if len(self.conversation_history) % 2 == 0:
            template = HumanMessagePromptTemplate.from_template(prompt)
        else:
            template = AIMessagePromptTemplate.from_template(prompt)

        self.conversation_history.append(template)

    def feedback(self, **kwargs):
        return LLMChain(
            llm=self.llm,
            prompt=ChatPromptTemplate.from_template(self.prompts["feedback"]),
        ).run(**kwargs)

    def refine(self):
        prompt = ChatPromptTemplate.from_messages(self.conversation_history)
        return LLMChain(llm=self.llm, prompt=prompt).run({})

    def isCompleted(self, res):
        return self.stop_kw in res

    def run(self, instruction, plan, result):
        self.addToHistory("instruction:" + instruction)
        self.addToHistory("plan:" + plan)
        self.addToHistory("code:" + result)
        for _ in trange(self.max_iter):
            feedback = self.feedback(instruction=instruction, result=result)
            self.addToHistory("feedback:" + feedback)
            if self.log_intermediate_steps:
                print(colored("feedback:\n" + feedback, "blue"))
            if self.isCompleted(feedback):
                return result
            if self.log_intermediate_steps:
                print(colored("refined result:\n" + result, "green"))
            result = self.refine().replace("{", "{{").replace("}", "}}")
            self.addToHistory("code:" + result)
        return result


if __name__ == "__main__":
    plan_refiner = SelfRefiner()
    instruction = """create a game where the system creates a story and stops at the exciting point and asks to the user 
    to make a selection then after user makes his selection, system continues to the story depending on the user's selection"""
    with open("~/Desktop/test/test.py") as f:
        result = f.read().replace("{", "{{").replace("}", "}}")
    plan_refiner.run(instruction=instruction, result=result)



import json
import os
import re
from difflib import SequenceMatcher

from langchain.chains import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (ChatPromptTemplate,
                                    HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)

from demogpt import utils
from demogpt.chains import prompts


class TaskChains:
    llm = None

    @classmethod
    def setLlm(
        cls,
        model,
        openai_api_key=os.getenv("OPENAI_API_KEY", ""),
        temperature=0.0,
        openai_api_base=None,
    ):
        cls.llm = ChatOpenAI(
            model=model,
            openai_api_key=openai_api_key,
            temperature=temperature,
            openai_api_base=openai_api_base,
        )

    @classmethod
    def getChain(cls, system_template="", human_template="", **kwargs):
        prompts = []
        if system_template:
            prompts.append(SystemMessagePromptTemplate.from_template(system_template))
        if human_template:
            prompts.append(HumanMessagePromptTemplate.from_template(human_template))
        chat_prompt = ChatPromptTemplate.from_messages(prompts)
        return LLMChain(llm=cls.llm, prompt=chat_prompt).run(**kwargs)

    @classmethod
    def uiInputText(cls, task):
        variable = ", ".join(task["output_key"])
        instruction = task["description"]
        code = cls.getChain(
            human_template=prompts.ui_input_text.human_template,
            instruction=instruction,
            variable=variable,
        )
        return utils.refine(code)

    @classmethod
    def uiOutputText(cls, task):
        args = ", ".join(task["input_key"])
        data_type = task["input_data_type"]
        if isinstance(args, list):
            args = ",".join(args)
        instruction = task["description"]
        code = cls.getChain(
            system_template=prompts.ui_output_text.system_template,
            human_template=prompts.ui_output_text.human_template,
            instruction=instruction,
            args=args,
            data_type=data_type,
        )
        return utils.refine(code)

    @classmethod
    def uiInputFile(cls, task):
        variable = ", ".join(task["output_key"])
        instruction = task["description"]
        res = cls.getChain(
            system_template=prompts.ui_input_file.system_template,
            human_template=prompts.ui_input_file.human_template,
            instruction=instruction
        )
        res = res[res.find("{") : res.rfind("}") + 1]
        res = json.loads(res)
        title = res.get("title")
        data_type = res.get("data_type")
        code = f"""
uploaded_file = st.file_uploader("{title}", type={data_type}, key='{variable}')
if uploaded_file is not None:
    # Create a temporary file to store the uploaded content
    extension = uploaded_file.name.split(".")[-1]
    with tempfile.NamedTemporaryFile(delete=False, suffix=f'.{{extension}}') as temp_file:
        temp_file.write(uploaded_file.read())
        {variable} = temp_file.name # it shows the file path
else:
    {variable} = ''
        """
        return code

    @classmethod
    def pathToContent(cls, task, code_snippets):
        instruction = task["description"]
        argument = ", ".join(task["input_key"])
        variable = ", ".join(task["output_key"])

        code = cls.getChain(
            system_template=prompts.path_to_file.system_template,
            human_template=prompts.path_to_file.human_template,
            instruction=instruction,
            argument=argument,
            variable=variable,
            code_snippets=code_snippets,
        )
        return utils.refine(code)

    @classmethod
    def promptTemplate(cls, task):
        inputs = ", ".join(task["input_key"])
        instruction = task["description"]

        res = cls.getChain(
            system_template=prompts.prompt_template.system_template,
            human_template=prompts.prompt_template.human_template,
            instruction=instruction,
            inputs=inputs,
        )
        res = res[res.find("{") : res.rfind("}") + 1]
        return json.loads(res)

    @classmethod
    def uiInputChat(cls, task):
        variable = ", ".join(task["output_key"])
        instruction = task["description"]

        code = cls.getChain(
            human_template=prompts.ui_input_chat.human_template,
            instruction=instruction,
            variable=variable,
        )
        return utils.refine(code)

    @classmethod
    def uiOutputChat(cls, task):
        res = ", ".join(task["input_key"])

        code = f"""
import time

with st.chat_message("assistant"):
    message_placeholder = st.empty()
    full_response = ""
    # Simulate stream of response with milliseconds delay
    for chunk in {res}.split():
        full_response += chunk + " "
        time.sleep(0.05)
        # Add a blinking cursor to simulate typing
        message_placeholder.markdown(full_response + "▌")
    message_placeholder.markdown(full_response)
    # Add assistant response to chat history
    if full_response:
        st.session_state.messages.append({{"role": "assistant", "content": full_response}})        
        """
        return code

    @classmethod
    def chat(cls, task):
        inputs = ", ".join(task["input_key"])
        instruction = task["description"]

        res = cls.getChain(
            system_template=prompts.chat.system_template,
            human_template=prompts.chat.human_template,
            instruction=instruction,
            inputs=inputs,
        )
        res = res.replace("'''", '"""')
        print("chat:")
        print(res)
        res = res[res.find("{") : res.rfind("}") + 1]
        return json.loads(res, strict=False)

    @classmethod
    def promptTemplateRefiner(cls, templates, inputs, feedback):
        res = cls.getChain(
            system_template=prompts.prompt_chat_refiner.system_template,
            human_template=prompts.prompt_chat_refiner.human_template,
            templates=templates,
            feedback=feedback,
            inputs=inputs,
        )
        print("promptTemplateRefiner:")
        print(res)
        res = res[res.find("{") : res.rfind("}") + 1]
        return json.loads(res)
    
    @classmethod
    def search_chat(cls, task):
        argument = ", ".join(task["input_key"])
        variable = ", ".join(task["output_key"])
        function_name = task["task_name"]
        instruction = task["description"]

        res = cls.getChain(
            system_template=prompts.search.system_template,
            human_template=prompts.search.human_template,
            instruction=instruction,
            inputs=argument,
        )
        
        res = res.replace('"',"'")

        code = f"""
from langchain.agents import ConversationalChatAgent, AgentExecutor
from langchain.tools import DuckDuckGoSearchRun
from langchain.agents.tools import Tool
from langchain.chains import LLMMathChain
from langchain.chat_models import ChatOpenAI

def {function_name}({argument}):
    llm = ChatOpenAI(model_name="gpt-3.5-turbo-16k", openai_api_key=openai_api_key)
    llm_math_chain = LLMMathChain.from_llm(llm=llm, verbose=True)
    tools = [
        DuckDuckGoSearchRun(name="Search"),
        Tool(
            name="Calculator",
            func=llm_math_chain.run,
            description="useful for when you need to answer questions about math"
        )]
    chat_agent = ConversationalChatAgent.from_llm_and_tools(llm=llm, tools=tools)
    executor = AgentExecutor.from_agent_and_tools(
        agent=chat_agent,
        tools=tools,
        memory=memory,
        return_intermediate_steps=False,
        handle_parsing_errors=True,
    )
    return executor({argument})["response"]

if not openai_api_key.startswith('sk-'):
    st.warning('Please enter your OpenAI API key!', icon='⚠')
    {variable} = ""
elif {argument}:
    {variable} = {function_name}({argument})
else:
    {variable} = ''
  
        """
        
        return code

    @classmethod
    def search(cls, task):
        argument = ", ".join(task["input_key"])
        variable = ", ".join(task["output_key"])
        function_name = task["task_name"]
        instruction = task["description"]

        res = cls.getChain(
            system_template=prompts.search.system_template,
            human_template=prompts.search.human_template,
            instruction=instruction,
            inputs=argument,
        )
        
        res = res.replace('"',"'")

        code = f"""
from langchain.chat_models import ChatOpenAI
from langchain_experimental.plan_and_execute import PlanAndExecute, load_agent_executor, load_chat_planner
from langchain.llms import OpenAI
from langchain.tools import DuckDuckGoSearchRun
from langchain.agents.tools import Tool
from langchain.chains import LLMMathChain

def {function_name}({argument}):
    search_input = "{res}".format({argument}={argument})
    llm = OpenAI(openai_api_key=openai_api_key, temperature=0)
    llm_math_chain = LLMMathChain.from_llm(llm=llm, verbose=True)
    tools = [
        DuckDuckGoSearchRun(name="Search"),
        Tool(
            name="Calculator",
            func=llm_math_chain.run,
            description="useful for when you need to answer questions about math"
        ),
    ]
    model = ChatOpenAI(openai_api_key=openai_api_key, temperature=0, model_name="gpt-4")
    planner = load_chat_planner(model)
    executor = load_agent_executor(model, tools, verbose=True)
    agent = PlanAndExecute(planner=planner, executor=executor, verbose=True)
    try:
        with st.spinner('DemoGPT is working on it. It might take 1-2 minutes...'):
            return agent.run(search_input)
    except AuthenticationError:
        st.warning('This tool requires GPT-4. Please enter a key that has GPT-4 access', icon="⚠️")
        return ''
        
if not openai_api_key.startswith('sk-'):
    st.warning('Please enter your OpenAI API key!', icon='⚠')
    {variable} = ""
elif {argument}:
    {variable} = {function_name}({argument})
else:
    {variable} = ''
        """
        return code

    @classmethod
    def docLoad(cls, task, code_snippets):
        
        def get_most_similar_key(input_key, available_keys):
            # This function returns the most similar key from available_keys to the input_key.
            best_match = None
            best_ratio = 0
            for key in available_keys:
                ratio = SequenceMatcher(None, input_key, key).ratio()
                if ratio > best_ratio:
                    best_ratio = ratio
                    best_match = key
            return best_match
        
        type2loader = {
            "txt": "TextLoader",
            "docx":"UnstructuredWordDocumentLoader",
            "pdf":"UnstructuredPDFLoader",
            "pptx":"UnstructuredPowerPointLoader",
            "csv":"CSVLoader",
            "xlsx":"UnstructuredExcelLoader",
            "zip":"NotionDirectoryLoader",
            "online_pdf":"OnlinePDFLoader",
            "web":"WebBaseLoader",
            "xlsx":"UnstructuredExcelLoader",
            "youtube":"YoutubeLoader",  
        }
        
        loader2type = {type2loader[dtype]:dtype for dtype in type2loader}
        
        def getLoaderCall(data_type):
            if data_type in loader2type:
                data_type = loader2type[data_type]
            
            loader = type2loader.get(data_type)  # First, try to get the exact match
            if loader is None:
                # If there's no exact match, get the most similar key and retrieve the value
                similar_key = get_most_similar_key(data_type, type2loader.keys())
                loader = type2loader[similar_key]
                
            loader = type2loader[data_type]
            
            if data_type in [
                "txt",
                "online_pdf",
                "docx",
                "csv"
            ]:
                loader_line = f"loader = {loader}({argument})"
            elif data_type == "web":
                loader_line = f"loader = {loader}([{argument}])"
            elif data_type in ["pdf", "pptx"]:
                loader_line = (
                    f'loader = {loader}({argument}, mode="elements", strategy="fast")'
                )
            elif data_type  == "xlsx":
                loader_line = f'loader = {loader}({argument}, mode="elements")'
            elif data_type == "youtube":
                loader_line = (
                    f"loader = {loader}.from_youtube_url({argument}, add_video_info=False)"
                )
            elif data_type == "zip":
                loader_line = f"""if os.path.exists('Notion_DB') and os.path.isdir('Notion_DB'):
            shutil.rmtree('Notion_DB')
        os.system(f"unzip {{{argument}}} -d Notion_DB")
        loader = {loader}("Notion_DB")"""
            else:
                loader_line = f"loader = TextLoader({argument})"
            
            return loader_line
        
        instruction = task["description"]
        argument = task["input_key"][0]
        variable = ", ".join(task["output_key"])
        function_name = task["task_name"]
        
        variable_match = re.search(r"(\w+)\s*=\s*temp_file\.name", code_snippets, re.MULTILINE)
        
        if variable_match:
            variable_name = variable_match.group(1).strip()
        else:
            variable_name = ''
            
        match = re.search(r"st\.file_uploader\(\s*?.*?type=\s*\[(.*?)\]\s*?.*?\)", code_snippets, re.DOTALL)
                
        loader_line = ""

        if variable_name == argument and  match:
            types = match.group(1).replace("'", "").split(", ")
            types = [type.strip() for type in types]
            if len(types) == 1:
                loader_line = getLoaderCall(types[0])
            else:
                loader_line = "\n    ".join([f"if {argument}.endswith('.{data_type}'):\n\t{getLoaderCall(data_type)}" for data_type in types])
        else:
            print("No match found. Using the chain...")
            types = []
            loader = cls.getChain(
                system_template=prompts.doc_load.system_template,
                human_template=prompts.doc_load.human_template,
                instruction=instruction,
                code_snippets=code_snippets
                )
            
            loader_line = getLoaderCall(loader)
            
        code = f"""
import shutil
from langchain.document_loaders import *

def {function_name}({argument}):
    {loader_line}
    docs = loader.load()
    return docs
if {argument}:
    {variable} = {function_name}({argument})
else:
    {variable} = ''
        """
        return code

    @classmethod
    def stringToDoc(cls, task):
        argument = ", ".join(task["input_key"])
        variable = ", ".join(task["output_key"])
        code = f"""
from langchain.docstore.document import Document
{variable} =  [Document(page_content={argument}, metadata={{'source': 'local'}})]
        """
        return code

    @classmethod
    def docToString(cls, task):
        argument = ", ".join(task["input_key"])
        variable = ", ".join(task["output_key"])
        code = f'{variable} = "".join([doc.page_content for doc in {argument}])'
        return code

    @classmethod
    def summarize(cls, task):
        argument = ", ".join(task["input_key"])
        variable = ", ".join(task["output_key"])
        function_name = task["task_name"]

        code = f"""
from langchain.chat_models import ChatOpenAI
from langchain.chains.summarize import load_summarize_chain

def {function_name}(docs):
    llm = ChatOpenAI(temperature=0, model_name="gpt-3.5-turbo-16k")
    chain = load_summarize_chain(llm, chain_type="stuff")
    with st.spinner('DemoGPT is working on it. It might take 5-10 seconds...'):
        return chain.run(docs)
if not openai_api_key.startswith('sk-'):
    st.warning('Please enter your OpenAI API key!', icon='⚠')
    {variable} = ""
elif {argument}:
    {variable} = summarize(argument)
else:
    variable = ""
"""
        return code

    @classmethod
    def pythonCoder(cls, task, code_snippets):
        instruction = task["description"]
        argument = ", ".join(task["input_key"])
        variable = ", ".join(task["output_key"])
        function_name = task["task_name"]

        code = cls.getChain(
            system_template=prompts.python_coder.system_template,
            human_template=prompts.python_coder.human_template,
            instruction=instruction,
            argument=argument,
            variable=variable,
            function_name=function_name,
            code_snippets=code_snippets,
        )
        return utils.refine(code)


human_template = """
variable: url
instruction: Get website URL from the user
code:
if url := st.chat_input("Enter website URL"):
    with st.chat_message("user"):
        st.markdown(url)
    st.session_state.messages.append({{"role": "user", "content": url}})

variable: source_text
instruction: Get source text from the user
code:
if source_text := st.chat_input("Type the source text"):
    with st.chat_message("user"):
        st.markdown(source_text)
    st.session_state.messages.append({{"role": "user", "content": source_text}})

variable: input_language
instruction: Get the input language from the user
code:
if input_language := st.chat_input("Enter the input language"):
    with st.chat_message("user"):
        st.markdown(input_language)
    st.session_state.messages.append({{"role": "user", "content": input_language}})

variable: {variable}
instruction: {instruction}
code:

"""



system_template = """
You are a coding assistant specialized in working with Streamlit applications and error-free code generation. 
Your task is, paying special attention to user input handling, state management and not getting any "not defined" error because of if statements without else. 
Generate nothing else but only the code so that it can be directly used.
"""

human_template = """
Directly repeat import statements and function definitions
When you define a function with if statement, put else and initialize it otherwise, you will get "not defined" error

Create a button to trigger the functionality of the app.

Don't touch imports, and function definitions but only manage the state of the application.
Call functions if all user inputs(if any) are taken and the button is pressed.

Original Code: 
{code_snippets}
################################
Error-free Code:
"""



system_template = """
Use the templates to guide the model in executing specific tasks or instructions. They act as directives, providing the context and structure needed for the model to respond appropriately.

Components:
1. "system_template": Describes the model's role and task for a given instruction. This string will be used with system_template.format(...) so only used curly braces for inputs
2. "template": Specifies the format for the model's response. This string will be used with template.format(...) so only used curly braces for inputs
3. "variety": Indicates how creative or deterministic the model's response should be.
4. "function_name": A unique identifier for the specific task or instruction.

IMPORTANT NOTE:
- Write "system_template" and "template" in a way that, (system_template+template).format(input=something for input in inputs) work.
What I mean is that, put all the elements of Inputs inside of either template or system_template with curly braces so that I can format it with predefined parameters.
"""

human_template = """
IMPORTANT NOTE:
- ONLY the variables listed under "Inputs" MUST be included in either the "system_template" or "template" section within curly braces (e.g., '{{variable_name}}'). Do NOT include any other parameters within curly braces.
- Ensure that the exact variable names listed in "Inputs" are used without any modifications.
- If a variable is listed in "Inputs," it must appear within curly braces in at least one of the "system_template" or "template" sections.
=========================================
Instruction: Generate a blog post from a title.
Inputs: ["title"]
Args: {{
"system_template": "You are an assistant designed to write a blog post from the given title: '{{title}}'.",
"template": "Title: {{title}}. Please compose a blog post based on this title.",
"variety": "True",
"function_name": "blogger"
}}
##########################################
Instruction: Implement a language translation app from one language to another.
Inputs: ["source_language","output_language", "text"]
Args: {{
"system_template": "You are a language translator. Your task is to translate text from {{source_language}} to {{output_language}}.",
"template": "Please translate the following text to {{output_language}}: '{{text}}'.",
"variety": "False",
"function_name": "translator"
}}
##########################################
Instruction: Generate an appropriate name for an animal.
Inputs: ["animal"]
Args: {{
"system_template": "You are tasked with creating a name for an animal. You generate concise and fitting names.",
"template": "The animal is a {{animal}}. Please create a good name for it.",
"variety": "True",
"function_name": "animalNameGenerator"
}}
##########################################
Instruction: Create a programming-related humor machine.
Inputs: []
Args: {{
"system_template": "You are designed to generate humor related to programming. Be creative and entertaining.",
"template": "Please generate a programming-related joke or humorous statement.",
"variety": "True",
"function_name": "humorGenerator"
}}
##########################################
Instruction: Act as a math teacher to solve a problem.
Inputs: ["math_problem"]
Args: {{
"system_template": "You are a virtual math teacher, capable of solving any given math problem.",
"template": "The problem is: {{math_problem}}. Please solve it and show the steps.",
"variety": "False",
"function_name": "mathSolver"
}}
##########################################
Instruction: Compose a piece of classical music.
Inputs: ["instrumentation", "theme"]
Args: {{
"system_template": "You are a composer creating a piece of classical music with specified instrumentation and theme.",
"template": "Compose a piece using the following instrumentation: {{instrumentation}}, based on the theme: '{{theme}}'.",
"variety": "True",
"function_name": "musicComposer"
}}
##########################################
Instruction:{instruction}
Inputs:{inputs}
Args:
"""



system_template = """
You are good at writing Python code.
You are supposed to create a function and call that function which does
the given instruction.
Here is the part of the code that you are supposed to continue:
{code_snippets}
"""

human_template = """
Write a function to load the file from the path for the argument name, variable and instruction below and also check if the path is not empty:
Instruction:{instruction}
Argument Name : {argument}
Variable Name : {variable}
Python Code:
"""



system_template = """
You are an AI assistant that write a concise prompt to direct an assistant to make web search for the given instruction.
You will have inputs and instruction. The prompt should be formattable with the inputs which means it should include inputs with curly braces.
"""

human_template = """
Instruction: Search the given input
Inputs:input
Prompt: Find the answer of it: {{input}}

Instruction: Find the list of song releated to the title
Inputs:title
Prompt: Find the list of songs releated to the title: {{title}}

Instruction:{instruction}
Inputs:{inputs}
Prompt:
"""



import logging
import os
import signal
import sys

import streamlit as st

import streamlit.components.v1 as components

current_file_path = os.path.abspath(__file__)
current_directory = os.path.dirname(current_file_path)
parent_directory = os.path.dirname(current_directory)
grandparent_directory = os.path.dirname(parent_directory)
sys.path.append(grandparent_directory)

from model import DemoGPT
from utils import runStreamlit


try:
    from dotenv import load_dotenv

    load_dotenv()
except Exception as e:
    logging.error("dotenv import error but no needed")


def generate_response(txt, title):
    """
    Generate response using the LangChainCoder.

    Args:
        txt (str): The input text.

    Yields:
        dict: A dictionary containing response information.
    """
    for data in agent(txt, title):
        yield data


def initCode():
    if "code" not in st.session_state:
        st.session_state["code"] = ""
        st.session_state.edit_mode = False


# Page title
title = "🧩 DemoGPT"

st.set_page_config(page_title=title)
    
st.title(title)


initCode()

# Text input

openai_api_key = st.sidebar.text_input(
    "OpenAI API Key",
    placeholder="sk-...",
    value=os.getenv("OPENAI_API_KEY", ""),
    type="password",
)

openai_api_base = st.sidebar.text_input(
    "Open AI base URL",
    placeholder="https://api.openai.com/v1",
)

models = (
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-16k-0613",
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-0613",
)

model_name = st.sidebar.selectbox("Model", models)

empty_idea = st.empty()
demo_idea = empty_idea.text_area(
    "Enter your LLM-based demo idea",
    placeholder="Type your demo idea here",
    height=100,
    help="""## Example prompts
* Character Clone: Want an app that converses like Jeff Bezos? Prompt - "Create me a chat-based application that talks like Jeff Bezos."
* Language Mastery: Need help in learning French? Prompt - "Create me an application that translates English sentences to French and provides pronunciation guidance for learners. 
* Content Generation: Looking to generate content? Prompt - "Create a system that can write ready to share Medium article from website. The resulting Medium article should be creative and interesting and written in a markdown format."
    """,
)

empty_title = st.empty()
demo_title = empty_title.text_input(
    "Give a name for your application",
    placeholder="Title",
    help="It will be displayed as a title in your app",
)


def progressBar(percentage, bar=None):
    if bar:
        bar.progress(percentage)
    else:
        return st.progress(percentage)


if "pid" not in st.session_state:
    st.session_state["pid"] = -1

if "done" not in st.session_state:
    st.session_state["done"] = False

with st.form("a", clear_on_submit=True):
    submitted = st.form_submit_button("Submit")


def kill():
    if st.session_state["pid"] != -1:
        logging.info(f"Terminating the previous applicaton ...")
        try:
            os.kill(st.session_state["pid"], signal.SIGTERM)
        except Exception as e:
            pass
        st.session_state["pid"] = -1


if submitted:
    st.session_state.messages = []
    if not openai_api_key:
        st.warning("Please enter your OpenAI API Key!", icon="⚠️")
    else:
        bar = progressBar(0)
        st.session_state.container = st.container()
        try:
            agent = DemoGPT(openai_api_key=openai_api_key, openai_api_base=openai_api_base)
            agent.setModel(model_name)
        except Exception as e:
            st.warning(e)
        else:
            kill()
            code_empty = st.empty()
            st.session_state.container = st.container()
            for data in generate_response(demo_idea, demo_title):
                done = data.get("done", False)
                failed = data.get("failed", False)
                message = data.get("message", "")
                st.session_state["message"] = message
                stage = data.get("stage", "stage")
                code = data.get("code", "")
                progressBar(data["percentage"], bar)

                st.session_state["done"] = done
                st.session_state["failed"] = failed
                st.session_state["message"] = message

                if done or failed:
                    st.session_state.code = code
                    break

                st.info(message, icon="🧩")
                st.session_state.messages.append(message)

elif "messages" in st.session_state:
    for message in st.session_state.messages:
        st.info(message, icon="🧩")

if st.session_state.done:
    st.success(st.session_state.message)
    with st.expander("Code", expanded=True):
        code_empty = st.empty()
        if st.session_state.edit_mode:
            new_code = code_empty.text_area("", st.session_state.code, height=500)
            if st.button("Save & Rerun"):
                st.session_state.code = (
                    new_code  # Save the edited code to session state
                )
                st.session_state.edit_mode = False  # Exit edit mode
                code_empty.code(new_code)
                kill()
                st.session_state["pid"] = runStreamlit(
                    new_code, openai_api_key, openai_api_base
                )
                st.experimental_rerun()

        else:
            code_empty.code(st.session_state.code)
            if st.button("Edit"):
                st.session_state.edit_mode = True  # Enter edit mode
                st.experimental_rerun()
    example_submitted = False
    if submitted:
        st.session_state["pid"] = runStreamlit(code, openai_api_key, openai_api_base)

if st.session_state.get("failed", False):
    with st.form("fail"):
        st.warning(st.session_state["message"])
        email = st.text_input("Email", placeholder="example@example.com")
        email_submit = st.form_submit_button("Send")
    if email_submit:
        st.success(
            "🌟 Thank you for entrusting us with your vision! We're on it and will ping you the moment your app is ready to launch. Stay tuned for a stellar update soon!"
        )


import os
import sys
from time import sleep
import openai

import autopep8
from tqdm import trange

from demogpt.chains.chains import Chains
from demogpt.chains.task_chains import TaskChains
from demogpt.utils import getCodeSnippet, getFunctionNames, init, reorderTasksForChatApp
import streamlit as st
import re


class DemoGPT:
    def __init__(
        self,
        openai_api_key=os.getenv("OPENAI_API_KEY", ""),
        model_name="gpt-3.5-turbo-0613",
        max_steps=10,
        openai_api_base="",
    ):
        assert len(
            openai_api_key.strip()
        ), "Either give openai_api_key as an argument or put it in the environment variable"
        self.model_name = model_name
        self.openai_api_key = openai_api_key
        self.max_steps = max_steps  # max iteration for refining the model purpose
        self.openai_api_base = openai_api_base
        self.FAIL_MESSAGE = """🚀✨ Impressive! While DemoGPT can handle a galaxy of app ideas, 
        you've shot for the stars with a unique one. We're ramping up our engines to meet such visionary requests. 
        Give us a little time, and we'll be right there with you.\n\n📧
        Care to leave your email? We'll notify you when we're ready for this stellar journey!"""
        
        self.gpt4_message = """
🚫 Access Denied to GPT-4!

Hey there! It looks like you're trying to create an app with GPT-4, but unfortunately, you don't have the required access. Fear not! You can:

Upgrade your access through OpenAI's platform.
Opt for another model that you have access to and give it another whirl.
We appreciate your understanding and look forward to seeing what you create! 😊
            """
        try:
            self.available_models = [model["id"] for model in openai.Model.list(api_key=openai_api_key)["data"] if "gpt" in model["id"]]
        except:
            self.available_models = []
            raise AssertionError("Invalid OpenAI API Key")
        else:
            assert (not self.model_name) or self.model_name in self.available_models , self.gpt4_message
                
        Chains.setLlm(
            self.model_name, self.openai_api_key, openai_api_base=self.openai_api_base, has_gpt4=self.hasGPT4
        )
        TaskChains.setLlm(
            self.model_name, self.openai_api_key, openai_api_base=self.openai_api_base
        )
                    
    @property
    def hasGPT4(self):
        return "gpt-4-0613" in self.available_models
                
    def setModel(self, model_name):
        if model_name not in self.available_models:
            raise AssertionError(self.gpt4_message)
        self.model_name = model_name
        Chains.setLlm(
            self.model_name, self.openai_api_key, openai_api_base=self.openai_api_base, has_gpt4=self.hasGPT4
        )
        TaskChains.setLlm(
            self.model_name, self.openai_api_key, openai_api_base=self.openai_api_base
        )

    def __repr__(self) -> str:
        return f"DemoGPT(model_name='{self.model_name}',max_steps={self.max_steps})"

    def __call__(
        self,
        instruction="Create a translation system that converts English to French",
        title="",
    ):
        
        yield {
            "stage": "system_inputs",
            "completed": False,
            "percentage": 0,
            "done": False,
            "message": "System inputs are being detected...",
            "failed": False,
        }

        system_inputs = Chains.systemInputs(instruction=instruction)
        app_type = Chains.appType(instruction=instruction)

        yield {
            "stage": "plan",
            "completed": False,
            "percentage": 10,
            "done": False,
            "message": "Plan creation has started...",
            "failed": False,
        }

        plan = Chains.planWithInputs(
            instruction=instruction, system_inputs=system_inputs, app_type=app_type
        )

        yield {
            "stage": "plan",
            "completed": True,
            "percentage": 20,
            "done": False,
            "message": "Plan has been generated.",
            "failed": False,
        }

        sleep(1)

        yield {
            "stage": "plan_controlling",
            "completed": True,
            "percentage": 25,
            "done": False,
            "message": "Plan is being controlled.",
            "failed": False,
        }

        plan_controller_result = Chains.planController(plan=plan, app_type=app_type)

        for _ in trange(self.max_steps//2):
            if not plan_controller_result["valid"]:
                plan = Chains.planRefiner(
                    instruction=instruction,
                    plan=plan,
                    feedback=plan_controller_result["feedback"],
                    app_type=app_type,
                )
                plan_controller_result = Chains.planController(
                    plan=plan, app_type=app_type
                )
            else:
                break

        yield {
            "stage": "task",
            "completed": False,
            "percentage": 30,
            "done": False,
            "message": "Task generation has started...",
            "failed": False,
        }

        task_list = Chains.tasks(
            instruction=instruction, plan=plan, app_type=app_type
        )

        yield {
            "stage": "task",
            "completed": True,
            "percentage": 50,
            "done": False,
            "message": "Tasks have been generated.",
            "tasks": task_list,
            "failed": False,
        }

        sleep(1)

        yield {
            "stage": "task_controlling",
            "completed": True,
            "percentage": 55,
            "done": False,
            "message": "Tasks are being controlled.",
            "failed": False,
        }

        task_controller_result = Chains.taskController(
            tasks=task_list, app_type=app_type
        )

        for _ in trange(self.max_steps):
            if not task_controller_result["valid"]:
                try:
                    task_list = Chains.refineTasks(
                        instruction=instruction,
                        tasks=task_list,
                        feedback=task_controller_result["feedback"],
                        app_type=app_type,
                    )
                    task_controller_result = Chains.taskController(
                        tasks=task_list, app_type=app_type
                    )
                except Exception as e:
                    print(e)
                    if "16k" in Chains.model:
                        break
                    st.toast(
                        "To increase the window size, changing model type to gpt-3.5-turbo-16k-0613"
                    )
                    Chains.setModel("gpt-3.5-turbo-16k-0613")
            else:
                break

        if not task_controller_result["valid"]:

            yield {
                "stage": "task",
                "completed": False,
                "percentage": 100,
                "done": False,
                "message": self.FAIL_MESSAGE,
                "failed": True,
            }

        else:
            
            title = Chains.title(instruction=instruction)
            
            task_list = reorderTasksForChatApp(task_list) # for chat apps, remove the code between chat input and chat output

            code_snippets = init(title)

            sleep(1)

            yield {
                "stage": "draft",
                "completed": False,
                "percentage": 60,
                "done": False,
                "message": "Converting tasks to code snippets...",
            }

            num_of_tasks = len(task_list)

            for i, task in enumerate(task_list):
                code = getCodeSnippet(task, code_snippets, self.max_steps)
                code = "#" + task["description"] + "\n" + code
                code_snippets += code
                yield {
                    "stage": "draft",
                    "completed": i + 1 == num_of_tasks,
                    "percentage": 60 + int(20 * (i + 1) / num_of_tasks),
                    "done": False,
                    "message": f"{i+1}/{num_of_tasks} tasks have been converted to code",
                    "code": code,
                }

            sleep(1)

            yield {
                "stage": "draft",
                "completed": False,
                "percentage": 85,
                "done": False,
                "message": "Code snippets are being combined...",
            }

            chat_app = any(
                [
                    task["task_type"] in ["ui_input_chat", "ui_output_chat", "chat"]
                    for task in task_list
                ]
            )

            if chat_app:
                final_code = code_snippets
                sleep(1)
            else:
                function_names = getFunctionNames(code_snippets)
                draft_code = Chains.combine_v2(
                    code_snippets=code_snippets, function_names=function_names
                )
                import_statements = Chains.imports(code_snippets=code_snippets)
                extra = ''
                if f"st.title('{title}')" not in draft_code:
                    extra = f"\nst.title('{title}')\n"
                if extra:
                    draft_code = extra + draft_code
                final_code = import_statements + draft_code

            # finalize the format
            final_code = autopep8.fix_code(final_code)
            
            how_to_markdown = Chains.howToUse(code_snippets=final_code)
            sleep(1)
            about = Chains.about(instruction=instruction, title=title)
            pattern = r'(openai_api_key\s*=\s*st\.sidebar\.text_input\((?:[^()]*|\([^)]*\))*\))'
            # replacement string with additional code
            replacement = how_to_markdown + r'\1' + about
            # substitute using regex
            final_code = re.sub(pattern, replacement, final_code, flags=re.DOTALL)

            yield {
                "stage": "final",
                "completed": True,
                "percentage": 100,
                "done": True,
                "message": "Final code has been generated. Directing to the demo page...",
                "code": final_code,
            }



#!/usr/bin/env python

import os
import subprocess


def main():
    current_dir = os.path.dirname(os.path.realpath(__file__))
    subprocess.run(["streamlit", "run", os.path.join(current_dir, "app.py")])


if __name__ == "__main__":
    main()



system_template = """
You are helpful assistant that can read code, feedcback and instruction to generate the refined version of the code
You are supposed to fix all the problems given to you in the feedback
"""

human_template = """
Instruction:{instruction}
Code:{code}
Feedback:{feedback}
Refined Code:
"""
